import json
from typing import Literal, Optional, Sequence, cast

import numpy as np
import pandas as pd
from typing_extensions import TypeGuard

import snowflake.snowpark
import snowflake.snowpark.functions as F
import snowflake.snowpark.types as spt
from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
)
from snowflake.ml._internal.utils import identifier
from snowflake.ml.model import type_hints as model_types
from snowflake.ml.model._deploy_client.warehouse import infer_template
from snowflake.ml.model._signatures import base_handler, core, pandas_handler


class SnowparkDataFrameHandler(base_handler.BaseDataHandler[snowflake.snowpark.DataFrame]):
    @staticmethod
    def can_handle(data: model_types.SupportedDataType) -> TypeGuard[snowflake.snowpark.DataFrame]:
        return isinstance(data, snowflake.snowpark.DataFrame)

    @staticmethod
    def count(data: snowflake.snowpark.DataFrame) -> int:
        return data.count()

    @staticmethod
    def truncate(data: snowflake.snowpark.DataFrame) -> snowflake.snowpark.DataFrame:
        return cast(snowflake.snowpark.DataFrame, data.limit(SnowparkDataFrameHandler.SIG_INFER_ROWS_COUNT_LIMIT))

    @staticmethod
    def validate(data: snowflake.snowpark.DataFrame) -> None:
        schema = data.schema
        for field in schema.fields:
            data_type = field.datatype
            try:
                core.DataType.from_snowpark_type(data_type)
            except snowml_exceptions.SnowflakeMLException:
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.INVALID_DATA,
                    original_exception=ValueError(
                        f"Data Validation Error: Unsupported data type {field.datatype} in column {field.name}."
                    ),
                )

    @staticmethod
    def infer_signature(
        data: snowflake.snowpark.DataFrame, role: Literal["input", "output"]
    ) -> Sequence[core.BaseFeatureSpec]:
        return pandas_handler.PandasDataFrameHandler.infer_signature(
            SnowparkDataFrameHandler.convert_to_df(data), role=role
        )

    @staticmethod
    def convert_to_df(
        data: snowflake.snowpark.DataFrame,
        ensure_serializable: bool = True,
        features: Optional[Sequence[core.BaseFeatureSpec]] = None,
    ) -> pd.DataFrame:
        # This method do things on top of to_pandas, to make sure the local dataframe got is in correct shape.
        dtype_map = {}
        if features:
            for feature in features:
                if isinstance(feature, core.FeatureGroupSpec):
                    raise snowml_exceptions.SnowflakeMLException(
                        error_code=error_codes.NOT_IMPLEMENTED,
                        original_exception=NotImplementedError("FeatureGroupSpec is not supported."),
                    )
                assert isinstance(feature, core.FeatureSpec), "Invalid feature kind."
                dtype_map[feature.name] = feature.as_dtype()
        df_local = data.to_pandas()
        # This is because Array will become string (Even though the correct schema is set)
        # and object will become variant type and requires an additional loads
        # to get correct data otherwise it would be string.
        for field in data.schema.fields:
            if isinstance(field.datatype, spt.ArrayType):
                df_local[identifier.get_unescaped_names(field.name)] = df_local[
                    identifier.get_unescaped_names(field.name)
                ].map(json.loads)
        # Only when the feature is not from inference, we are confident to do the type casting.
        # Otherwise, dtype_map will be empty
        df_local = df_local.astype(dtype=dtype_map)
        return df_local

    @staticmethod
    def convert_from_df(
        session: snowflake.snowpark.Session, df: pd.DataFrame, keep_order: bool = False
    ) -> snowflake.snowpark.DataFrame:
        # This method is necessary to create the Snowpark Dataframe in correct schema.
        # However, in this case, the order could not be preserved. Thus, a _ID column has to be added,
        # if keep_order is True.
        # Although in this case, the column with array type can get correct ARRAY type, however, the element
        # type is not preserved, and will become string type. This affect the implementation of convert_from_df.
        df = pandas_handler.PandasDataFrameHandler.convert_to_df(df)
        df_cols = df.columns
        if df_cols.dtype != np.object_:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.NOT_IMPLEMENTED,
                original_exception=ValueError("Cannot convert a Pandas DataFrame whose column index is not a string"),
            )
        features = pandas_handler.PandasDataFrameHandler.infer_signature(df, role="input")
        # Role will be no effect on the column index. That is to say, the feature name is the actual column name.
        sp_df = session.create_dataframe(df)
        column_names = []
        columns = []
        for feature in features:
            if isinstance(feature, core.FeatureGroupSpec):
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.NOT_IMPLEMENTED,
                    original_exception=NotImplementedError("FeatureGroupSpec is not supported."),
                )
            assert isinstance(feature, core.FeatureSpec), "Invalid feature kind."
            column_names.append(identifier.get_inferred_name(feature.name))
            columns.append(F.col(identifier.get_inferred_name(feature.name)).cast(feature.as_snowpark_type()))

        sp_df = sp_df.with_columns(column_names, columns)

        if keep_order:
            sp_df = sp_df.with_column(infer_template._KEEP_ORDER_COL_NAME, F.monotonically_increasing_id())

        return sp_df



from abc import abstractmethod
from typing import Protocol, final

from snowflake.ml._internal import migrator_utils
from snowflake.ml.model._packager.model_meta import model_meta


class _BaseModelHandlerMigratorProtocol(Protocol):
    source_version: str
    target_version: str

    @staticmethod
    @abstractmethod
    def upgrade(
        name: str,
        model_meta: model_meta.ModelMetadata,
        model_blobs_dir_path: str,
    ) -> None:
        raise NotImplementedError


class BaseModelHandlerMigrator(_BaseModelHandlerMigratorProtocol):
    @final
    def try_upgrade(self, name: str, model_meta: model_meta.ModelMetadata, model_blobs_dir_path: str) -> None:
        assert (
            model_meta.models[name].handler_version == self.__class__.source_version
        ), "Incorrect source handler version found."
        try:
            self.upgrade(name=name, model_meta=model_meta, model_blobs_dir_path=model_blobs_dir_path)
            model_meta.models[name].handler_version = self.__class__.target_version
        except migrator_utils.UnableToUpgradeError as e:
            raise RuntimeError(
                f"Can not upgrade your model {name} from version {self.__class__.source_version} to"
                f" {self.__class__.target_version}."
                f"The latest version support the original version of Snowpark ML library is {e.last_supported_version}."
            )



from abc import abstractmethod
from typing import Dict, Generic, Optional, Protocol, Type, final

from typing_extensions import TypeGuard, Unpack

from snowflake.ml.model import custom_model, type_hints as model_types
from snowflake.ml.model._packager.model_handlers_migrator import base_migrator
from snowflake.ml.model._packager.model_meta import model_meta


class _BaseModelHandlerProtocol(Protocol[model_types._ModelType]):
    HANDLER_TYPE: model_types.SupportedModelHandlerType
    HANDLER_VERSION: str
    _MIN_SNOWPARK_ML_VERSION: str
    _HANDLER_MIGRATOR_PLANS: Dict[str, Type[base_migrator.BaseModelHandlerMigrator]]

    @classmethod
    @abstractmethod
    def can_handle(cls, model: model_types.SupportedDataType) -> TypeGuard[model_types._ModelType]:
        """Whether this handler could support the type of the `model`.

        Args:
            model: The model object.

        Raises:
            NotImplementedError: Not Implemented
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def cast_model(cls, model: model_types.SupportedModelType) -> model_types._ModelType:
        """Cast the model from Union type into the type that handler could handle.

        Args:
            model: The model object.

        Raises:
            NotImplementedError: Not Implemented
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def save_model(
        cls,
        name: str,
        model: model_types._ModelType,
        model_meta: model_meta.ModelMetadata,
        model_blobs_dir_path: str,
        sample_input: Optional[model_types.SupportedDataType] = None,
        is_sub_model: Optional[bool] = False,
        **kwargs: Unpack[model_types.BaseModelSaveOption],
    ) -> None:
        """Save the model.

        Args:
            name: Name of the model.
            model: The model object.
            model_meta: The model metadata.
            model_blobs_dir_path: Directory path to the model.
            sample_input: Sample input to infer the signatures from.
            is_sub_model: Flag to show if it is a sub model, a sub model does not need signature.
            kwargs: Additional saving options.

        Raises:
            NotImplementedError: Not Implemented
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def load_model(
        cls,
        name: str,
        model_meta: model_meta.ModelMetadata,
        model_blobs_dir_path: str,
        **kwargs: Unpack[model_types.ModelLoadOption],
    ) -> model_types._ModelType:
        """Load the model into memory.

        Args:
            name: Name of the model.
            model_meta: The model metadata.
            model_blobs_dir_path: Directory path to the whole model.
            kwargs: Options when loading the model.

        Raises:
            NotImplementedError: Not Implemented
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def convert_as_custom_model(
        cls,
        raw_model: model_types._ModelType,
        model_meta: model_meta.ModelMetadata,
        **kwargs: Unpack[model_types.ModelLoadOption],
    ) -> custom_model.CustomModel:
        """Create a custom model class wrap for unified interface when being deployed. The predict method will be
        re-targeted based on target_method metadata.

        Args:
            raw_model: original model object,
            model_meta: The model metadata.
            kwargs: Options when converting the model.

        Raises:
            NotImplementedError: Not Implemented
        """
        raise NotImplementedError


class BaseModelHandler(Generic[model_types._ModelType], _BaseModelHandlerProtocol[model_types._ModelType]):
    """
    Provides handling for a given type of model defined by `HANDLER_TYPE` class property.

    HANDLER_TYPE: The string type that identify the handler. Should be unique in the library.
    HANDLER_VERSION: The version of the handler.
    _MIN_SNOWPARK_ML_VERSION: The minimal version of Snowpark ML library to use the current handler.
    _HANDLER_MIGRATOR_PLANS: Dict holding handler migrator plans.

    MODELE_BLOB_FILE_OR_DIR: Relative path of the model blob file in the model subdir. Default to "model.pkl".
    MODEL_ARTIFACTS_DIR: Relative path of the model artifacts dir in the model subdir. Default to "artifacts"
    DEFAULT_TARGET_METHODS: Default target methods to be logged if not specified in this kind of model. Default to
        ["predict"]
    IS_AUTO_SIGNATURE: Set to True if the model could get model signature automatically and do not require user
        inputting sample data or model signature. Default to False.
    """

    MODELE_BLOB_FILE_OR_DIR = "model.pkl"
    MODEL_ARTIFACTS_DIR = "artifacts"
    DEFAULT_TARGET_METHODS = ["predict"]
    IS_AUTO_SIGNATURE = False

    @classmethod
    @final
    def try_upgrade(cls, name: str, model_meta: model_meta.ModelMetadata, model_blobs_dir_path: str) -> None:
        """Try upgrade the stored model to adapt latest handler

        Args:
            name: Name of the model.
            model_meta: The model metadata.
            model_blobs_dir_path: Directory path to the whole model.

        Raises:
            RuntimeError: Raised when there is no corresponding migrator available.
        """
        while model_meta.models[name].handler_version != cls.HANDLER_VERSION:
            if model_meta.models[name].handler_version not in cls._HANDLER_MIGRATOR_PLANS.keys():
                raise RuntimeError(
                    f"Can not find migrator to migrate model {name} from {model_meta.models[name].handler_version}"
                    f" to version {cls.HANDLER_VERSION}."
                )
            migrator = cls._HANDLER_MIGRATOR_PLANS[model_meta.models[name].handler_version]()
            migrator.try_upgrade(
                name=name,
                model_meta=model_meta,
                model_blobs_dir_path=model_blobs_dir_path,
            )
